#!/usr/bin/perl
#-------------------------------------------------------------------------------
#	$Id: bilderwichtel.pl,v 1.22 2023/02/14 21:26:35 wolf Exp $
#-------------------------------------------------------------------------------
#
#	Glob funktioniert auch bei Blanks im Dateinamen
#	Hinweis-Zeilen werden bei Programmende entfernt.
#
#-------------------------------------------------------------------------------
#
#	Copyright (c) 2002  Wolfgang Stanglmeier, Siegburg, Germany
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
#
#   3. All advertising materials mentioning features or use of this
#      software must display the following acknowledgement:
#
#        This product includes software developed by
#        Wolfgang Stanglmeier, Siegburg, Germany
#
#   4. The name of the author may not used to endorse or promote
#      products derived from this software without specific prior
#      written permission.
#
#  This software is provided by the author ``AS IS'' and any express
#  or implied warranties, including, but not limited to, the implied
#  warranties of merchantibility and fitness for a particular purpose
#  are disclaimed.
#  In no event shall the author be liable for any direct, indirect,
#  incidental, special, exemplary or consequential damages (including,
#  but not limited to, procurement or substitute goods or services;
#  loss of use, data, or profits; or business interruption) however
#  caused and on any theory of liablility, whether in contract, strict
#  liability, or tort (including negligence or otherwise) arising in
#  any way out of the use of this software, even if advised of the
#  possibility of such damage.
#
#-------------------------------------------------------------------------------
#
#       $Log: bilderwichtel.pl,v $
#       Revision 1.22  2023/02/14 21:26:35  wolf
#       + Feldname unique
#
#       Revision 1.21  2023/02/12 14:04:05  wolf
#       + Glob mit Leerzeichen
#       + Macro "THEME"
#       * Schnips ohne gemeinsame scaled Temp-Datei
#       + auch Schnips nicht neugebaut wenn existierend.
#
#       Revision 1.20  2023/01/31 05:13:12  wolf
#       + sub scaleImage
#       + sub scaleAndCutImage
#       + use scaleImage, scaleAndCutImage
#       + scaleAndCutImage: use GD
#
#       Revision 1.19  2023/01/31 03:37:08  wolf
#       * USE_GD -> USE_NETPBM
#
#       Revision 1.18  2023/01/30 21:38:16  wolf
#       + rcode patcht auch Schnips-Tabellen
#
#       Revision 1.17  2023/01/30 20:36:03  wolf
#       + automatische Konversion von CP1252
#       + akzeptiert Bilder mit falscher Groß/Klein-Schreibung durch Globbing
#       + ignoriere changelogfile
#
#       Revision 1.16  2023/01/05 16:12:09  wolf
#       + remove <!--! xxx !--> style comments
#
#       Revision 1.15  2022/12/27 15:56:43  wolf
#       + log longtext files
#
#       Revision 1.14  2022/12/22 22:56:29  wolf
#       !!! fix regexps
#       * handle undefs
#
#       Revision 1.13  2022/12/22 20:08:17  wolf
#       + handling undefined (FATAL => all)
#       !!! bugfix makro name ohne wert
#
#       Revision 1.12  2022/12/22 02:56:04  wolf
#       + Leerzeichen in Macro-Def erhalten
#       + DestUriStoreField zentral löschen
#       + tc-Funktion
#
#       Revision 1.11  2022/12/20 22:06:57  wolf
#       + JPEG_QUALITY
#
#       Revision 1.10  2022/12/20 19:33:49  wolf
#       +++ use GD
#
#       Revision 1.9  2022/12/20 11:46:42  wolf
#       *** empty log message ***
#
#       Revision 1.8  2022/12/20 10:41:50  wolf
#       * layout
#       * $DatabaseFile -> @DatabaseFiles
#
#       Revision 1.7  2022/12/20 09:50:57  wolf
#       * use keys %FieldDefined
#
#       Revision 1.6  2022/12/20 09:01:15  wolf
#       + komplett incl schnips
#
#       Revision 1.5  2022/12/20 04:03:09  wolf
#       + computeImageSize
#
#       Revision 1.4  2022/12/20 02:03:10  wolf
#       *** GenPage
#
#       Revision 1.3  2022/12/17 03:50:55  wolf
#       *** empty log message ***
#
#       Revision 1.2  2022/12/15 22:19:20  wolf
#       *** empty log message ***
#
#       Revision 1.1  2022/12/15 05:38:21  wolf
#       Initial revision
#
#       Revision 1.8  2009/06/15 18:40:21  xinu
#       *** empty log message ***
#
#       Revision 1.7  2009/01/19 23:48:39  wolf
#       *** empty log message ***
#
#       Revision 1.6  2009/01/18 21:57:32  wolf
#       *** empty log message ***
#
#       Revision 1.5  2009/01/13 18:50:51  wolf
#       *** empty log message ***
#
#	Revision 1.4  2009/01/13 18:47:14  wolf
#	Kleinere Aenderungen
#
#       Revision 1.18  2005/02/07 21:12:31  sabine
#       *** empty log message ***
#
#       Revision 1.17  2004/06/06 08:14:11  sabine
#       *** empty log message ***
#
#       Revision 1.3  2003/03/30 16:15:55  wolf
#       Filenamefilter für Texte.
#
#       Revision 1.2  2003/03/30 15:38:44  wolf
#       Jetzt auch mit Langtexten.
#
#       Revision 1.1  2003/03/24 20:54:38  wolf
#       Initial revision
#
#-------------------------------------------------------------------------------

use strict;
use utf8;
use Encode;
use POSIX;

use warnings FATAL => "all";

binmode STDOUT, ':utf8';
binmode STDERR, ':utf8';

my $CLREOL = "\x1b[K";

#-------------------------------------------------------------------------------
#
#	configuration
#
#-------------------------------------------------------------------------------

my %Option;

while( @ARGV && $ARGV[ 0 ] =~ /^ -(\w+) (?: = (\S+) )? $/x ){

	$Option{ $1 } = $2 ? $2 : 1;
	shift( @ARGV );
}

my $Verbosity = int( $Option{ 'v' } // 0 );

#-------------------------------------------------------------------------------
#
#	error handling
#
#-------------------------------------------------------------------------------

our( $File, $Lino, $LineContent, $nErrors );

sub Info{
	my( $msg ) = @_;
	warn "$File:$Lino: $msg\n";
}

sub Fatal{
	&Error( @_ );
	&CheckErrors;
}

sub Error{
	my( $msg ) = @_;
	warn "$File:$Lino: $msg\n";
	$nErrors++;
}

sub CheckErrors{
	return unless $nErrors;
	die "$nErrors Errors, terminating.\n";
}

#-------------------------------------------------------------------------------
#
#	config file
#
#-------------------------------------------------------------------------------

unless( @ARGV && shift( @ARGV ) =~ m!^ (\S+) $!x ){

	die "No configuration file specified.\n";
}

my $ConfigFile = $1;

unless( -f $ConfigFile && -r $ConfigFile ){

	die "Configuration file '$ConfigFile' is not a redable file.\n";
}

my $BaseDir = &absname( &dirname( $ConfigFile ), getcwd );

#-------------------------------------------------------------------------------
#
#	Configuration values read from configuration file
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#	Pages and images are created in this directory and subdirectories
#-------------------------------------------------------------------------------

my $DestinationBase;

#-------------------------------------------------------------------------------
#	Csv-files containing image names and meta data
#-------------------------------------------------------------------------------

my @DatabaseFiles;

#-------------------------------------------------------------------------------
#	CSV-file field names
#-------------------------------------------------------------------------------

my %FieldNameToIndex;
my @DbFieldNames;
my @DbFieldFlags;

#-------------------------------------------------------------------------------
#	Database
#-------------------------------------------------------------------------------

my %Database;		# array[ entry, fields ] of data
my %FieldDefined;	# all columns
my @AllItems;		# all rows

#-------------------------------------------------------------------------------
#	Source images are read from these directories
#-------------------------------------------------------------------------------

my @ImageSourceDirectories;

#-------------------------------------------------------------------------------
#	Names long texts are read from these directories
#-------------------------------------------------------------------------------

my @TextSourceDirectories;

#-------------------------------------------------------------------------------
#	Layout templates are read from these directories
#-------------------------------------------------------------------------------

my @LayoutSourceDirectories;

#-------------------------------------------------------------------------------
#	Macros
#-------------------------------------------------------------------------------

my %Macro;
{
	my @t = localtime;
	$Macro{ 'DD'}   = sprintf( '%02u', $t[ 3 ] );
	$Macro{ 'MM'}   = sprintf( '%02u', $t[ 4 ] + 1 );
	$Macro{ 'YY'}   = sprintf( '%02u', $t[ 5 ] % 100 );
	$Macro{ 'YYYY'} = sprintf( '%04u', $t[ 5 ] + 1900 );
}

#-------------------------------------------------------------------------------
#	Image size constraints
#-------------------------------------------------------------------------------

my %ImageHeightLimit;
my %ImagePixelLimit;
my %ImageWidthLimit;

#-------------------------------------------------------------------------------
#	Items selected by last 'je' or 'foreach' directive
#-------------------------------------------------------------------------------

my @SelectedItems;

#-------------------------------------------------------------------------------
#	Virtual fields + patching
#-------------------------------------------------------------------------------

my $CurrentField;
my $CurrentExpression;

our $CurrentItem;
our $ActiveItem;

our %CurrentData;

#-------------------------------------------------------------------------------
#	Image and Page generation
#-------------------------------------------------------------------------------

my $DestUriExpression;
my $DestUri;
my $DestUriStoreField;

my $PageLayoutFile;

my $ImageSourceField;
my $ImageSizeId;

#-------------------------------------------------------------------------------
#	Long texts read from text file
#-------------------------------------------------------------------------------

my %LongTexts;

#-------------------------------------------------------------------------------
#	Open main config file
#-------------------------------------------------------------------------------

$nErrors = 0;

$File = $ConfigFile;
$Lino = 0;

unless( open( CONF, '<:raw', $ConfigFile ) ){

	die "Cannot open configuration file '$ConfigFile': $!\n";
}

#-------------------------------------------------------------------------------
#	Read and process main config file
#-------------------------------------------------------------------------------

while( <CONF> ){

	++$Lino;

	#-----------------------------------------------------------------------
	#	skip empty and comment lines
	#-----------------------------------------------------------------------

	next if     /^ \s* [#] /x;
	next unless /      \w  /x;

	s!\s+$!!;

	#-----------------------------------------------------------------------
	#	utf8 / iso decode
	#-----------------------------------------------------------------------

	$LineContent = decode_charset( $_ );

	#-----------------------------------------------------------------------
	#	be verbose
	#-----------------------------------------------------------------------

	warn "$File:$Lino:[ $LineContent ]\n" if $Verbosity >= 2;

	#-----------------------------------------------------------------------
	#	split line to command and fields
	#-----------------------------------------------------------------------

	my @args = split( /\s+/, $LineContent );
	my $cmd  = lc( shift( @args ));

	#-----------------------------------------------------------------------
	#	'zielverzeichnis'	{directory}
	#-----------------------------------------------------------------------

	if( $cmd eq 'zielverzeichnis' || $cmd eq 'destination_directory' ){

		&Fatal( "Destination directory multiple defined." )
							if $DestinationBase;
		&Fatal( "$cmd needs exactly one argument (directory)." )
							unless @args == 1;

		my $dir = &absname( shift( @args ));

		if( $dir =~ m!/$! ){

			&Error( "Destination directory '$dir' may not end with '/'." );
			next;
		}

		unless( -d $dir ){

			&Error( "Destination directory '$dir' is not a directory." );
			next;
		}

		unless( -w $dir ){
			&Error( "Destination directory '$dir' is not writable." );
			next;
		}

		$DestinationBase = $dir;
		$Macro{ 'THEME' } = genid( $dir =~ s!^.*/!!r );

		next;
	}

	#-----------------------------------------------------------------------
	#	'datenbankdatei'	{file}
	#-----------------------------------------------------------------------

	if( $cmd eq 'datenbankdatei' || $cmd eq 'database_file' ){

		&Fatal( "$cmd needs one argument (file)." ) unless @args;

		my $file = &absname( shift( @args ) );

		unless( -f $file ){
			&Error( "Database file '$file' is not a file." );
			next;
		}

		unless( -r $file ){
			&Error( "Database file '$file' is not readable." );
			next;
		}

		push( @DatabaseFiles, $file );
		next;
	}

	#-----------------------------------------------------------------------
	#	'bilderverzeichnis'	{directory1} {directory2} ...
	#-----------------------------------------------------------------------

	if( $cmd eq 'bilderverzeichnis' || $cmd eq 'image_directory' ){

		foreach my $arg( @args ){

			my $dir = &absname( $arg );

			unless( -d $dir ){
				&Error( "Image directory '$dir' is not a directory" );
				next;
			}

			unless( -x $dir ){
				&Error( "Image directory '$dir' is not searchable." );
				next;
			}

			push( @ImageSourceDirectories, $dir );
		}
		next;
	}

	#-----------------------------------------------------------------------
	#	'textverzeichnis'	{directory1} {directory2} ...
	#-----------------------------------------------------------------------

	if( $cmd eq 'textverzeichnis' || $cmd eq 'textfile_directory' ){

		foreach my $arg( @args ){

			my $dir = &absname( $arg );

			unless( -d $dir ){
				&Error( "Textfile directory '$dir' is not a directory" );
				next;
			}

			unless( -x $dir ){
				&Error( "Textfile directory '$dir' is not searchable." );
				next;
			}

			push( @TextSourceDirectories, $dir );
		}
		next;
	}

	#-----------------------------------------------------------------------
	#	'layoutverzeichnis'	{directory1} {directory2} ...
	#-----------------------------------------------------------------------

	if( $cmd eq 'layoutverzeichnis' || $cmd eq 'layout_directory'){

		foreach my $arg( @args ){

			my $dir = &absname( $arg );

			unless( -d $dir ){
				&Error( "Layout directory '$dir' is not a directory" );
				next;
			}

			unless( -x $dir ){
				&Error( "Layout directory '$dir' is not searchable." );
				next;
			}

			push( @LayoutSourceDirectories, $dir );
		}
		next;
	}

	#-----------------------------------------------------------------------
	#	'datenbankfeld'	{feld1} {feld2} ...
	#-----------------------------------------------------------------------

	if( $cmd eq 'datenbankfeld' || $cmd eq 'database_field' ){

		foreach my $arg( @args ){

			unless( $arg =~ /^ (\w+) (:unique)? $/x ){

				&Error( "Invalid Database field name '$arg' (\\w+)." );
				next;
			}

			push( @DbFieldFlags, $2 ? "u" : "" );

			$FieldNameToIndex{ $1 } = scalar( @DbFieldNames );
			push( @DbFieldNames, $1 );

			$FieldDefined{ $1 } = 1;
		}
		next;
	}

	#-----------------------------------------------------------------------
	#	'makro'	{name} {...wert...}
	#-----------------------------------------------------------------------

	if( $cmd eq 'makro' || $cmd eq 'macro' ){

		&Fatal( "$cmd needs argument (macro name)" ) unless @args;
		$_ = shift( @args );

		unless( /^ ( [\w\pL]+ ) $/x ){

			&Error( "Invalid macro name '$_' (\\w+)." );
			next;
		}

		my $name = $1;

		if( defined $Macro{ $name }){

			&Error( "Macro '$name' is multiple defined." );
			next;
		}

		#my $value = join( "\x20", @args );
		my $value = $LineContent =~ s!^\s*\S+\s+\S*\s*!!r;

		%CurrentData = %Macro;	# initialize CurrentData

		$Macro{ $name } = $value =~ s!^ " (.*) " $!$1!xr =~ s!\\n!\n!gr;
		next;
	}

	#-----------------------------------------------------------------------
	#	'bildgroesse'	{id} {max-width} {max-height} {max-pixels}
	#-----------------------------------------------------------------------

	if( $cmd eq 'bildgroesse' || $cmd eq 'image_size' ){

		&Fatal( "$cmd needs 4 args: id, max-width, max-height, max-n-pixels." )
							unless @args == 4;
		$_ = shift( @args );

		unless( /^ ( \w+ ) $/x ){

			&Error( "Invalid image size id '$_' (\\w+)." );
			next;
		}

		my $id = $1;

		my $w = int( $args[ 0 ] // 0 );
		my $h = int( $args[ 1 ] // 0 );
		my $p = int( $args[ 2 ] // 0 );

		$ImageWidthLimit { $id } = $w;
		$ImageHeightLimit{ $id } = $h;
		$ImagePixelLimit { $id } = $p;
		next;
	}

	#-----------------------------------------------------------------------
	#	'changelogfile'		-- removed --
	#-----------------------------------------------------------------------

	next if $cmd eq 'changelogfile';

	#-----------------------------------------------------------------------
	#	Weitere Kommandos nur mit kompletter Konfiguration
	#-----------------------------------------------------------------------

	&CheckErrors;
	&Error( "No destination directory defined." ) unless $DestinationBase;
	&Error( "No database files defined."        ) unless @DatabaseFiles;
	&Error( "No Image directory defined."       ) unless @ImageSourceDirectories;
	&Error( "No database field names defined."  ) unless @DbFieldNames;
	&CheckErrors;

	#-----------------------------------------------------------------------
	#	Read database files (once)
	#-----------------------------------------------------------------------

	unless( %Database ){

		my %fieldValDef;

		my $RequiredNValues = scalar( @DbFieldNames );

		local( $File, $Lino );

		foreach my $db_file( @DatabaseFiles ){

			$File = $db_file;
			$Lino = 0;

			my $ne = 0;

			#-------------------------------------------------------
			#	open database csv file
			#-------------------------------------------------------

			unless( open( IN, '<:raw', $File ) ){

				&Fatal( "openread '$File': $!\n" );
				next;
			}

			while( <IN> ){
				++$Lino;

				#-----------------------------------------------
				#	skip empty and comment lines
				#-----------------------------------------------

				next if     /^ \s* [#] /x;
				next unless /      \w  /x;

				#-----------------------------------------------
				#	normalize and split
				#-----------------------------------------------

				s!^\s+!!;
				s!\s+$!!;

				#-----------------------------------------------
				#	decode
				#-----------------------------------------------

				$_ = decode_charset( $_ );

				#-----------------------------------------------
				#	split
				#-----------------------------------------------

				my @values = split( / \s* [|] \s* /x, $_, -1 );
				next unless @values;

				#-----------------------------------------------
				#	check number of values
				#-----------------------------------------------

				my $nValues = scalar( @values );

				if( $nValues < $RequiredNValues ){

	&Error( "Too few fields: $RequiredNValues expected, $nValues found." );
					next;
				}

				if( $nValues > $RequiredNValues ){

	&Error( "Too many fields: $RequiredNValues expected, $nValues found." );
					next;
				}

				#-----------------------------------------------
				#	item id (limited to 9999 items)
				#-----------------------------------------------

				my $item = sprintf( '%04u', @AllItems + 1 );

				#-----------------------------------------------
				#	store field values
				#-----------------------------------------------

				for( my $i=0; $i < @DbFieldNames; $i++ ){

					my $fieldName  = $DbFieldNames[ $i ];
					my $fieldValue = $values[ $i ];

					$Database{ $item, $fieldName } =
								$fieldValue;

					if( $DbFieldFlags[ $i ] ){

						my $pos = $fieldValDef{ $fieldName, $fieldValue };

						if( $pos ){

							&Error( "\n$pos: field '$fieldName' value '$fieldValue' not unique" );
						} else {

							$fieldValDef{ $fieldName, $fieldValue } = "$File:$Lino";
						}
					}
				}

				push( @AllItems, $item );

				++$ne;
			}

			#-------------------------------------------------------
			#	close csv file
			#-------------------------------------------------------

			close( IN );

			#-------------------------------------------------------
			#	be verbose
			#-------------------------------------------------------

			warn "$File: $ne items.\n" if $Verbosity >= 1;
		}

		#---------------------------------------------------------------
		#	sanity
		#---------------------------------------------------------------

		&Error( "Database contains no items." ) unless %Database;
		&CheckErrors;

		#---------------------------------------------------------------
		#	be verbose
		#---------------------------------------------------------------

		my $n_items = scalar @AllItems;
		&Info( "Read $n_items items from database file '$File'." );

		#---------------------------------------------------------------
		#	initialize selection
		#---------------------------------------------------------------

		@SelectedItems = @AllItems;
		next;
	}

	#-----------------------------------------------------------------------
	#	'setze'	{name} {...expression...}
	#-----------------------------------------------------------------------

	if( $cmd eq 'setze' || $cmd eq 'define' ){

		#---------------------------------------------------------------
		#	name
		#---------------------------------------------------------------

		&Fatal( "$cmd needs one argument (field name)." ) unless @args;
		$_ = shift( @args );

		unless( /^ ( \w+ ) $/x ){

			&Error( "Invalid field name '$_' (\\w+)" );
			next;
		}

		$CurrentField = $1;

		#---------------------------------------------------------------
		#	definition expression
		#---------------------------------------------------------------

		$CurrentExpression = join( "\x20", @args );

		#---------------------------------------------------------------
		#	execute procedure on selected items
		#---------------------------------------------------------------

		&ProcessItems( 'SetVar', @SelectedItems );

		#---------------------------------------------------------------
		#	mark field name as used
		#---------------------------------------------------------------

		$FieldDefined{ $CurrentField } = 1;
		next;
	}

	#-----------------------------------------------------------------------
	#	'je'	{dbfeld1} {dbfeld2} ...
	#	'je'	*				-> select all
	#	'je'					-> select one (first)
	#-----------------------------------------------------------------------
	#	Auswahl von Datenbank-Zeilen
	#-----------------------------------------------------------------------

	if( $cmd eq 'je' || $cmd eq 'foreach' ){

		#---------------------------------------------------------------
		#	select all
		#---------------------------------------------------------------

		if( @args == 1 && $args[ 0 ] eq '*' ){

			@SelectedItems = @AllItems;

			warn "\nSelected: ALL\n\n" if $Verbosity >= 1;
			next;
		}

		#---------------------------------------------------------------
		#	gather selection fields -> @uf
		#---------------------------------------------------------------

		my @selFields;

		foreach my $arg( @args ){

			unless( defined $FieldDefined{ $arg }){

				&Error( "Database field '$arg' is not defined." );
				next;
			}
			push( @selFields, $arg );
		}

		#---------------------------------------------------------------
		#	no fields -> select one (first)
		#---------------------------------------------------------------

		if( !@selFields ){

			@SelectedItems = ( $AllItems[ 0 ] );
			warn "\nSelected: FIRST( @SelectedItems )\n\n"
							if $Verbosity >= 1;
			next;
		}

		#---------------------------------------------------------------
		#	clear selection
		#---------------------------------------------------------------

		warn "\nSelect-Fields: ", join( "\x20", @selFields ), "\n"
							if $Verbosity >= 1;

		@SelectedItems = ();

		#---------------------------------------------------------------
		#	process all items
		#---------------------------------------------------------------

		my %seen;
		foreach my $item( @AllItems ){

			#------------------------------------------------------
			#	join entry values for selected fields
			#------------------------------------------------------

			my @values;
			foreach my $selField( @selFields ){

				push( @values, $Database{ $item, $selField });
			}
			my $valueSet = join( $;, @values );

			#------------------------------------------------------
			#	be verbose
			#------------------------------------------------------

			if( $Verbosity >= 3 ){

				my $seen = $seen{ $valueSet } // "";
				warn "Value: $item -> $valueSet ($seen)\n";
			}

			#------------------------------------------------------
			#	add to item list first entry having value-set
			#------------------------------------------------------

			next if $seen{ $valueSet };
			$seen{ $valueSet } = $item;

			push( @SelectedItems, $item );
		}

		warn "Selected: ", join( "\x20", @SelectedItems ), "\n\n"
							if $Verbosity >= 1;
		next;
	}

	#-----------------------------------------------------------------------
	#	sort selected database items
	#-----------------------------------------------------------------------

	if( $cmd eq 'sort' ){

		#---------------------------------------------------------------
		#	gather sort fields
		#---------------------------------------------------------------

		my @sortFields;

		foreach my $arg( @args ){

			unless( defined $FieldDefined{ $arg} ){

				&Error( "Database field '$arg' is not defined." );
				next;
			}
			push( @sortFields, $arg );
		}

		warn "Sort fields: ", join( "\x20", @sortFields ), "\n"
							if $Verbosity >= 1;

		#---------------------------------------------------------------
		#	gather combined values per selected item
		#---------------------------------------------------------------

		my %sort;

		foreach my $item( @SelectedItems ){

			my @sortValues;

			foreach my $sortField( @sortFields ){

				push( @sortValues, $Database{ $item, $sortField} );
			}
			$sort{ $item } = join( $;, @sortValues );
		}

		@SelectedItems = sort{ $sort{$a} cmp $sort{$b}} @SelectedItems;
		warn "Sorted: ", join( "\x20", @SelectedItems ), "\n"
							if $Verbosity >= 1;
		next;
	}

	#-----------------------------------------------------------------------
	#	group database items
	#-----------------------------------------------------------------------

	if( $cmd eq 'gruppe' || $cmd eq 'group' ){

		#---------------------------------------------------------------
		#	group name (=field name)
		#---------------------------------------------------------------

		&Fatal( "$cmd braucht einen Feld-Namen als Argument" )
							unless @args;
		$_ = shift( @args );

		unless( /^ ( \w+ ) $/x ){

			&Error( "Feldname '$_' ist nicht erlaubt (nur a-z,0-9,_)" );
			next;
		}

		$CurrentField = $1;

		$FieldDefined{ $CurrentField } = 1;

		#---------------------------------------------------------------
		#	gather group fields
		#---------------------------------------------------------------

		my @groupFields;

		foreach my $arg( @args ){

			unless( defined $FieldDefined{ $arg} ){

				&Error( "Group field '$arg' is not defined." );
				next;
			}
			push( @groupFields, $arg );
		}
		warn "\nNeue Gruppe '$CurrentField' [@groupFields]\n"
							if $Verbosity >= 1;

		#---------------------------------------------------------------
		#	no group field: all items in same group
		#---------------------------------------------------------------

		if( !@groupFields ){

			my $itemList = join( "\x20", @SelectedItems );

			foreach my $item( @AllItems ){

				$Database{ $item, $CurrentField} = $itemList;
			}

			warn "Group: ALL[ $itemList ]\n" if $Verbosity >= 1;
			next;
		}

		#---------------------------------------------------------------
		#	gather items per joined values of group fields
		#---------------------------------------------------------------

		my %gruppe;

		foreach my $item( @SelectedItems ){

			#-------------------------------------------------------
			#	get value-set of group fields per item
			#-------------------------------------------------------

			my @values;
			foreach my $field( @groupFields ){
				push( @values, $Database{ $item, $field} );
			}
			my $valueSet = join( $;, @values );

			#-------------------------------------------------------
			#	add item to group for the value-sets
			#-------------------------------------------------------

			$gruppe{ $valueSet } .= "$item ";
		}

		#---------------------------------------------------------------
		#	gather items per joined values of group fields
		#---------------------------------------------------------------

		foreach my $valueSet( sort keys %gruppe ){

			my @items = grep( /\w/, split( /\s+/,
						$gruppe{ $valueSet }));
			my $itemList = join( "\x20", @items );

			warn "Group: $itemList [$valueSet]\n"
							if $Verbosity >= 1;

			foreach my $item( @items ){
				$Database{ $item, $CurrentField} = $itemList;
			}
		}
		next;
	}

	#-----------------------------------------------------------------------
	#	generate images
	#-----------------------------------------------------------------------

	if( $cmd eq 'bild' || $cmd eq 'image' ){

&Fatal( "$cmd needs 4 args: img source, dest uri, img size id, img url field" )
			unless @args == 4;

		( $ImageSourceField, $DestUriExpression, $ImageSizeId,
						$DestUriStoreField ) = @args;

		unless( defined $FieldNameToIndex{ $ImageSourceField }){

	&Error( "Image source field '$ImageSourceField' is not defined." );
			next;
		}

		unless( defined $ImageWidthLimit{ $ImageSizeId }){

	&Error( "Image size id '$ImageSizeId' is not defined." );
			next;
		}

		#---------------------------------------------------------------
		#	use GD module?
		#---------------------------------------------------------------

		if( !defined $Macro{ 'USE_NETPBM' } ){

			use GD;
			GD::Image->trueColor( 1 );
		}

		&ProcessItems( 'GenImage', @SelectedItems )
					unless $Option{ 'keep_images' };
		next;
	}

	#-----------------------------------------------------------------------
	#	generate split images
	#-----------------------------------------------------------------------

	if( $cmd eq 'schnips' ){

&Fatal( "$cmd needs 4 args: img source, dest uri, img size id, img url field" )
			unless @args == 4;

		( $ImageSourceField, $DestUriExpression, $ImageSizeId,
						$DestUriStoreField ) = @args;

		unless( defined $FieldNameToIndex{ $ImageSourceField }){

	&Error( "Image source field '$ImageSourceField' is not defined." );
			next;
		}

		unless( defined $ImageWidthLimit{ $ImageSizeId }){

	&Error( "Image size id '$ImageSizeId' is not defined." );
			next;
		}

		#---------------------------------------------------------------
		#	use GD module?
		#---------------------------------------------------------------

		if( !defined $Macro{ 'USE_NETPBM' } ){

			use GD;
			GD::Image->trueColor( 1 );
		}

		&ProcessItems( 'GenSchnips', @SelectedItems )
					unless $Option{ 'keep_images' };
		next;
	}

	#-----------------------------------------------------------------------
	#	generate page
	#-----------------------------------------------------------------------

	if( $cmd eq 'seite' || $cmd eq 'page' ){

		&Fatal( "$cmd needs 2 args: layout_file and destination_uri." )
							unless @args == 2;

		( $PageLayoutFile, $DestUriExpression ) = @args;

		&ProcessItems( 'GenPage', @SelectedItems );
		next;
	}

	#-----------------------------------------------------------------------
	#	unknown directive
	#-----------------------------------------------------------------------

	&Error( "Unknown directive '$cmd'\n>> $LineContent" );
}

&CheckErrors;

#-------------------------------------------------------------------------------
#
#	Done
#
#-------------------------------------------------------------------------------

print STDERR "\r$CLREOL";

warn "\nOk :-)\n\n";
exit 0;

#-------------------------------------------------------------------------------
#
#	Apply function to selected items
#
#-------------------------------------------------------------------------------

sub ProcessItems{

	my( $procedureRef, @items ) = @_;

	foreach my $item( @items ){

		$CurrentItem = $item;

		#---------------------------------------------------------------
		#	%CurrentData := %Macro | DataBase{ item }
		#---------------------------------------------------------------
		#	defined macros plus database row of current item
		#---------------------------------------------------------------

		local( %CurrentData );
		%CurrentData = %Macro;

		foreach my $f( sort keys %FieldDefined ){

			$CurrentData{ $f } = $Database{ $CurrentItem, $f };
		}

		#---------------------------------------------------------------
		#	execute function
		#---------------------------------------------------------------

		{
			no strict 'refs';
			&$procedureRef;
		}

		#---------------------------------------------------------------
		#	store created object's url or html
		#---------------------------------------------------------------

		if( $DestUriStoreField ){

	$Database{ $item, $DestUriStoreField      } = $DestUri;
	$Database{ $item, $DestUriStoreField.'_W' } = $CurrentData{ 'W' };
	$Database{ $item, $DestUriStoreField.'_H' } = $CurrentData{ 'H' };

	$FieldDefined{ $DestUriStoreField      }  = 1;
	$FieldDefined{ $DestUriStoreField.'_W' }  = 1;
	$FieldDefined{ $DestUriStoreField.'_H' }  = 1;
		}
	}

	$DestUriStoreField = undef;
}

#-------------------------------------------------------------------------------
#
#	ProcessItem: Set Variable
#
#-------------------------------------------------------------------------------

sub SetVar{

	$Database{ $CurrentItem, $CurrentField } = &patch( $CurrentExpression );
}

#-------------------------------------------------------------------------------
#
#	ProcessItem: Generate Image
#
#-------------------------------------------------------------------------------
#	(!) Assumes JPEG source image
#-------------------------------------------------------------------------------

sub GenImage{

	#-----------------------------------------------------------------------
	#	image name
	#-----------------------------------------------------------------------

	die unless $ImageSourceField;

	my $name = $CurrentData{ $ImageSourceField };
	unless( $name ){

		&Error( "Source field '$ImageSourceField' is empty." );
		return;
	}

	#-----------------------------------------------------------------------
	#	lookup source image and compute source and destination size
	#-----------------------------------------------------------------------

	my $infile = computeImageSize( $name );

	my $w  = $CurrentData{ 'W' };
	my $h  = $CurrentData{ 'H' };

	#-----------------------------------------------------------------------
	#	destination relative path
	#-----------------------------------------------------------------------

	$DestUri = &patch( $DestUriExpression );

	#-----------------------------------------------------------------------
	#	destination path (absolute or relative to current directory)
	#-----------------------------------------------------------------------

	my $dest_file = "$DestinationBase/$DestUri";

	#-----------------------------------------------------------------------
	#	if dest file exists, recreate only if option -f
	#-----------------------------------------------------------------------

	if( -f $dest_file ){

		print STDERR "\r$DestUri: already exists.$CLREOL";
		return unless $Option{ 'f' };
	}

	#-----------------------------------------------------------------------
	#	create output directory
	#-----------------------------------------------------------------------

	&create_path( $dest_file );
	return if $nErrors;

	#-----------------------------------------------------------------------
	#	scale into temp file
	#-----------------------------------------------------------------------

	my $temp_file = "$dest_file.$$.tmp";

	scaleImage( $infile, $w, $h, $temp_file );

	#-----------------------------------------------------------------------
	#	rename tempfile to destination file
	#-----------------------------------------------------------------------

	&rename_if_changed( $temp_file, $dest_file );
}

#-------------------------------------------------------------------------------
#
#	Scale image
#
#-------------------------------------------------------------------------------

sub scaleImage {

	#-----------------------------------------------------------------------
	#	scale image file, using NETPBM or GD
	#-----------------------------------------------------------------------

	my( $srcFile, $dstWidth, $dstHeight, $dstFile ) = @_;

	#-----------------------------------------------------------------------
	#	use LIBJPEG and NETPBM
	#-----------------------------------------------------------------------

	if( defined $Macro{ 'USE_NETPBM' } ){

		my $cmd =
"djpeg '$srcFile' | pnmscale -width '$dstWidth' -height '$dstHeight' | cjpeg > '$dstFile'";
		warn "$cmd ...\n" if $Verbosity >= 3;

		system $cmd;

		if( $? ){

			&Error( "Pipe [$cmd] failed: status=$?." );
			unlink( $dstFile );
			return;
		}
		return;
	}

	#-----------------------------------------------------------------------
	#	use GD
	#-----------------------------------------------------------------------

	my $srcWidth  = $CurrentData{ 'OW' };
	my $srcHeight = $CurrentData{ 'OH' };

	my $srcImage  = GD::Image->new( $srcFile );
	my $dstImage  = GD::Image->new( $dstWidth, $dstHeight );

	#---------------------------------------------------------------
	#	$dstImage -> copyResampled( $srcImage,
	#		$dstX,     $dstY,      $srcX,     $srcY,
	#		$dstWidth, $dstHeight, $srcWidth, $srcHeight )
	#---------------------------------------------------------------

	$dstImage->copyResampled( $srcImage,
		0,         0,          0,         0,
		$dstWidth, $dstHeight, $srcWidth, $srcHeight );

	#---------------------------------------------------------------
	#	write to dst file
	#---------------------------------------------------------------

	unless( open( OUT, '>:raw', $dstFile ) ){

		&Error( "create '$dstFile': $!" );
		unlink( $dstFile );
		return undef;
	}

	my $quality = $Macro{ 'JPEG_QUALITY' };

	print OUT defined($quality) ?
		$dstImage -> jpeg( $quality ) :
		$dstImage -> jpeg();

	unless( close( OUT ) ){

		&Error( "close '$dstFile': $!");
		unlink( $dstFile );
		return undef;
	}
}

#-------------------------------------------------------------------------------
#
#	ProcessItem: Generate Image split to 5 parts
#
#-------------------------------------------------------------------------------
#	(!) Assumes JPEG source image
#-------------------------------------------------------------------------------
#
#		+----------------+-------+
#		|                |       |
#		|        a       |       |
#		|                |       |
#		+--------+-------+   b   +
#		|        |       |       |
#		|        |   d   |       |
#		|        |       |       |
#		|   c    +-------+-------+
#		|        |               |
#		|        |       e       |
#		|        |               |
#		+--------+---------------+
#
#-------------------------------------------------------------------------------

sub GenSchnips{

	#-----------------------------------------------------------------------
	#	image name
	#-----------------------------------------------------------------------

	die unless $ImageSourceField;

	my $name = $CurrentData{ $ImageSourceField };
	unless( $name ){

		&Error( "Source field '$ImageSourceField' is empty." );
		return;
	}

	#-----------------------------------------------------------------------
	#	lookup source image and compute source and destination size
	#-----------------------------------------------------------------------

	my $infile = computeImageSize( $name );

	return unless $infile;

	my $W  = $CurrentData{ 'W' };
	my $H  = $CurrentData{ 'H' };

	#-----------------------------------------------------------------------
	#	compute split positions
	#-----------------------------------------------------------------------

	my( %px, %py, %pw, %ph );

	my $x1 = int( 0.276 * $W );
	my $x2 = int( 0.724 * $W );
	my $y1 = int( 0.276 * $H );
	my $y2 = int( 0.724 * $H );

	$px{ 'a' } = 0;
	$pw{ 'a' } = $x2;
	$py{ 'a' } = 0;
	$ph{ 'a' } = $y1;

	$px{ 'b' } = $x2;
	$pw{ 'b' } = $W - $x2;
	$py{ 'b' } = 0;
	$ph{ 'b' } = $y2;

	$px{ 'c' } = 0;
	$pw{ 'c' } = $x1;
	$py{ 'c' } = $y1;
	$ph{ 'c' } = $H - $y1;

	$px{ 'd' } = $x1;
	$pw{ 'd' } = $x2 - $x1;
	$py{ 'd' } = $y1;
	$ph{ 'd' } = $y2 - $y1;

	$px{ 'e' } = $x1;
	$pw{ 'e' } = $W - $x1;
	$py{ 'e' } = $y2;
	$ph{ 'e' } = $H - $y2;

	#-----------------------------------------------------------------------
	#	create image parts
	#-----------------------------------------------------------------------

	my %img;

	foreach my $part( sort keys %px ){

		#---------------------------------------------------------------
		#	split parameter
		#---------------------------------------------------------------

		$CurrentData{ '_part_' } = $part;

		my $x = $CurrentData{ '_x_' } = $px{ $part };
		my $y = $CurrentData{ '_y_' } = $py{ $part };
		my $w = $CurrentData{ '_w_' } = $pw{ $part };
		my $h = $CurrentData{ '_h_' } = $ph{ $part };

		#---------------------------------------------------------------
		#	destination relative path
		#---------------------------------------------------------------

		$DestUri = &patch( $DestUriExpression );

		#---------------------------------------------------------------
		#	destination path (abs. or relative to current directory)
		#---------------------------------------------------------------

		my $dest_file = "$DestinationBase/$DestUri";

		#---------------------------------------------------------------
		#	image access html
		#---------------------------------------------------------------

		$img{ $part }=
	"<img src=\"${DestUri}\" width=\"$w\" height=\"$h\" alt=\"\">";

		#---------------------------------------------------------------
		#	if dest file exists, recreate only if option -f
		#---------------------------------------------------------------

		if( -f $dest_file ){

			print STDERR "\r$DestUri: already exists.$CLREOL";
			next unless $Option{ 'f' };
		}

		#---------------------------------------------------------------
		#	create output directory
		#---------------------------------------------------------------

		&create_path( $dest_file );

		if( $nErrors ){

			unlink( $dest_file );
			return undef;
		}

		#---------------------------------------------------------------
		#	create destination image part
		#---------------------------------------------------------------

		my $temp_file = "$dest_file.$$.tmp";

		#---------------------------------------------------------------
		#	scaleAndCutImage(
		#		$srcFile,
		#		$dstWidth, $dstHeight,
		#		$cutX, $cutY, $cutWidth, $cutHeight,
		#		$dstFile )
		#---------------------------------------------------------------

		scaleAndCutImage( $infile, $W, $H, $x, $y, $w, $h, $temp_file );

		return unless -s $temp_file;

		&rename_if_changed( $temp_file, $dest_file );
	}

	#-----------------------------------------------------------------------
	#	create combined image parts html
	#-----------------------------------------------------------------------
	#
	#	+----------------+-------+
	#	|                |       |
	#	|        a       |       |
	#	|                |       |
	#	+--------+-------+   b   +
	#	|        |       |       |
	#	|        |   d   |       |
	#	|        |       |       |
	#	|   c    +-------+-------+
	#	|        |               |
	#	|        |       e       |
	#	|        |               |
	#	+--------+---------------+
	#
	#-----------------------------------------------------------------------

	$DestUri = <<"EOT";
<table class="schnips">
<tr>
<td colspan="2">${img{'a'}}</td>
<td rowspan="2">${img{'b'}}</td>
</tr><tr>
<td rowspan="2">${img{'c'}}</td>
<td>${img{'d'}}</td>
</tr><tr>
<td colspan="2">${img{'e'}}</td>
</tr></table>
EOT
}

#-------------------------------------------------------------------------------
#
#	Scale and cut image
#
#-------------------------------------------------------------------------------

sub scaleAndCutImage {

	#-----------------------------------------------------------------------
	#	scale image file, than extract given rectangle
	#-----------------------------------------------------------------------

	my( $srcFile, $dstWidth, $dstHeight,
		$cutX, $cutY, $cutWidth, $cutHeight, $dstFile ) = @_;

	#-----------------------------------------------------------------------
	#	use
	#-----------------------------------------------------------------------

	if( defined $Macro{ 'USE_NETPBM' } ){

		my $quality = $Macro{ 'JPEG_QUALITY' } // 75;
		my $cmd = <<"EOT" =~ s!\s+!\x20!gr;
djpeg '$srcFile' |
pnmscale -width '$dstWidth' -height '$dstHeight' |
pnmcut $cutX $cutY $cutWidth $cutHeight |
cjpeg -qual '$quality' >$dstFile
EOT
		#---------------------------------------------------------------
		#	execute command
		#---------------------------------------------------------------

		#warn "$cmd ...\n";
		system $cmd;

		if( $? ){
			&Error( "Pipe [$cmd] failed': status=$?." );
			unlink( $dstFile );
			return;
		}
	}

	#-----------------------------------------------------------------------
	#	$srcFile,
	#	$dstWidth, $dstHeight,
	#	$cutX, $cutY, $cutWidth, $cutHeight,
	#	$dstFile
	#-----------------------------------------------------------------------

	my $srcWidth  = $CurrentData{ 'OW' };
	my $srcHeight = $CurrentData{ 'OH' };

	my $srcImage  = GD::Image->new( $srcFile );
	my $tmpImage  = GD::Image->new( $dstWidth, $dstHeight );
	my $dstImage  = GD::Image->new( $cutWidth, $cutHeight );

	#---------------------------------------------------------------
	#	scale to intermediate image
	#---------------------------------------------------------------
	#	copyResampled( $srcImage,
	#		$dstX,     $dstY,      $srcX,     $srcY,
	#		$dstWidth, $dstHeight, $srcWidth, $srcHeight )
	#---------------------------------------------------------------

	$tmpImage->copyResampled( $srcImage,
		0,         0,          0,         0,
		$dstWidth, $dstHeight, $srcWidth, $srcHeight );

	#---------------------------------------------------------------
	#	cut selected rectangle
	#---------------------------------------------------------------
	#	$dstImage -> copy( $srcImage,
	#		$dstX,	$dstY,	$srcX,	$srcY,
	#		$width,	$height)
	#---------------------------------------------------------------

	$dstImage->copy( $tmpImage,
		0,         0,          $cutX,     $cutY,
		$cutWidth, $cutHeight );

	#---------------------------------------------------------------
	#	write to dst file
	#---------------------------------------------------------------

	unless( open( OUT, '>:raw', $dstFile ) ){

		&Error( "create '$dstFile': $!" );
		unlink( $dstFile );
		return undef;
	}

	my $quality = $Macro{ 'JPEG_QUALITY' };

	print OUT defined($quality) ?
		$dstImage -> jpeg( $quality ) :
		$dstImage -> jpeg();

	unless( close( OUT ) ){

		&Error( "close '$dstFile': $!");
		unlink( $dstFile );
		return undef;
	}
}

#-------------------------------------------------------------------------------
#
#	get source and destination size of image
#
#-------------------------------------------------------------------------------

sub computeImageSize{

	#-----------------------------------------------------------------------
	#	image name
	#-----------------------------------------------------------------------

	my( $name ) = @_;

	unless( $name ){

		&Error( "computeImageSize: missing image name." );
		return undef;
	}

	#-----------------------------------------------------------------------
	#	image file
	#-----------------------------------------------------------------------

	my $infile;

	my $uncasedName = $name =~ s!([a-z])!'['.uc($1).lc($1).']'!eigr;

	foreach my $dir( @ImageSourceDirectories ){

		#---------------------------------------------------------------
		#	direct lookup
		#---------------------------------------------------------------

		$infile = "$dir/$name";
		last if -f $infile && -r $infile;
		$infile = '';

		#---------------------------------------------------------------
		#	no case lookup using glob
		#---------------------------------------------------------------

		my @glob = sort glob qq{ "$dir/$uncasedName" };
		next unless @glob;

		if( @glob >= 2 ){

			my $nGlobs = scalar@glob;

			&Error( "Glob($name) liefert $nGlobs Treffer:" );
			foreach my $glob( @glob ){
				&Info( "\x20\x20\x20\x20\x20\x20\x20\x20($glob)" );
			}
			return undef;
		}
			
		$infile = $glob[ 0 ];
		&Info( "Globbed: $infile" ) if $Verbosity >= 1;
		last if -f $infile && -r $infile;
		$infile = '';
	}

	unless( $infile ){

		&Error( "Image '$name' not found." );
		return undef;
	}

	#-----------------------------------------------------------------------
	#	extract orig size
	#-----------------------------------------------------------------------

	my( $ow, $oh ) = getImageSize( $infile );

	return undef unless $ow && $oh;

	$CurrentData{ 'OW' } = $ow;
	$CurrentData{ 'OH' } = $oh;

	#-----------------------------------------------------------------------
	#	compute destination size
	#-----------------------------------------------------------------------

	my $minScale = 1.0;

	#-----------------------------------------------------------------------
	#	max width
	#-----------------------------------------------------------------------

	my $max_w = $ImageWidthLimit{ $ImageSizeId };

	if( $max_w > 0 ){

		my $scale = $max_w / $ow ;
		if( $scale < $minScale ){ $minScale = $scale; }
	}

	#-----------------------------------------------------------------------
	#	max height
	#-----------------------------------------------------------------------

	my $max_h = $ImageHeightLimit{ $ImageSizeId };

	if( $max_h > 0 ){

		my $scale = $max_h / $oh;
		if( $scale < $minScale ){ $minScale = $scale; }
	}

	#-----------------------------------------------------------------------
	#	max n pixels
	#-----------------------------------------------------------------------

	my $max_p = $ImagePixelLimit{ $ImageSizeId };

	if( $max_p>0 ){

		my $scale = sqrt( $max_p / ($oh*$ow) );
		if( $scale < $minScale ){ $minScale = $scale; }
	}

	#-----------------------------------------------------------------------
	#	store destination size
	#-----------------------------------------------------------------------

	my $w = $CurrentData{ 'W' } = POSIX::round( $minScale * $ow );
	my $h = $CurrentData{ 'H' } = POSIX::round( $minScale * $oh );

	#warn "[ $ow x $oh ] → [ $w x $h ]\n";

	return $infile;
}

#-------------------------------------------------------------------------------
#
#	get width and height of image
#
#-------------------------------------------------------------------------------

sub getImageSize {

	my( $infile ) = @_;

	#-----------------------------------------------------------------------
	#	JPEG-Size
	#-----------------------------------------------------------------------

	my( $w, $h ) = getJpegSize( $infile, 0 );

	return( $w, $h ) if $w && $h;

	#-----------------------------------------------------------------------
	#	use NETPBM?
	#-----------------------------------------------------------------------

	if( defined $Macro{ 'USE_NETPBM' } ){

		#---------------------------------------------------------------
		#	use NETPBM
		#---------------------------------------------------------------

		my $cmd = "djpeg '$infile'";

		unless( open( PIPE, '-|', $cmd ) ){

			&Error( "openpipe [$cmd]: $!" );
			return undef;
		}

		$_ = <PIPE>;				# P[1-6] (type)

		unless( m!^ P ([1-6])\b !x ){

			&Error( "Invalid image '$infile' (no jpeg?)" );
			return undef;
		}

		$_ = <PIPE>;				# (width) (height)
		unless( m!^ \s* ( \d+ ) \s+ ( \d+ ) \b !x ){

			&Error( "Invalid image '$infile' (no jpeg?)" );
			return;
		}

		close( PIPE );

		$CurrentData{ 'OW' } = $w = $1;
		$CurrentData{ 'OH' } = $h = $2;

		return( $w, $h );
	}

	#-----------------------------------------------------------------------
	#	use GD
	#-----------------------------------------------------------------------

	use GD;

	#-----------------------------------------------------------------------
	#	any image size
	#-----------------------------------------------------------------------

	( $w, $h ) = GD::Image->new( $infile )->getBounds();

	return( $w, $h ) if $w && $h;

	#-----------------------------------------------------------------------
	#	failed
	#-----------------------------------------------------------------------

	&Error( "Unsupported image file '$infile'." );
	return undef;
}

#-------------------------------------------------------------------------------
#
#	Get jpeg file's width and height
#
#-------------------------------------------------------------------------------

sub getJpegSize {

	my( $filename, $debug ) = @_;
	warn "getJpegSize: file '$filename'\n" if $debug;

	return undef unless $filename =~ m! \. jpe?g $!ix;

	open( JPEGSIZE, '<:raw', $filename ) || return undef;
	warn "getJpegSize: Jpeg-Header?\n" if $debug;

	read( JPEGSIZE, $_, 2 );
	return undef unless $_ eq "\xff\xd8";

	my $offset = 2;
	for( ;; ){

		return undef unless read( JPEGSIZE, $_, 4 ) == 4;
		my( $mark, $tag, $size ) = unpack 'CCn', $_;

		warn sprintf "getJpegSize: Segment \@%04x: [mark=%02x,tag=%02x,size=%04x]...\n",
			$offset, $mark, $tag, $size if $debug;

		return undef unless $mark == 0xff;
		return undef unless $size >= 2;

		$size   -= 2;
		$offset += 4;

		return undef unless read( JPEGSIZE, $_, $size ) == $size;

		$offset += $size;

		next unless $tag == 0xc0 && $size >= 5;
		my( undef, $h, $w ) = unpack 'Cnn', $_;

		warn "getJpegSize: $filename: $w x $h\n" if $debug;
		return( $w, $h );
	}
}

#-------------------------------------------------------------------------------
#
#	ProcessItem: Gen Page
#
#-------------------------------------------------------------------------------

sub GenPage{

	#-----------------------------------------------------------------------
	#	lookup layout template in list of template directorues
	#-----------------------------------------------------------------------

	my $name = $PageLayoutFile;
	die unless $name;

	my $infile;

	foreach my $dir( @LayoutSourceDirectories ){

		$infile = "$dir/$name";
		last if -f $infile && -r $infile;
		$infile='';
	}

	unless( $infile ){

		&Error( "Layout file '$name' not found." );
		return;
	}

	#-----------------------------------------------------------------------
	#	read layout template file
	#-----------------------------------------------------------------------

	my $page_source;

	open( my $fh, '<:raw', $infile ) or die "openread '$infile': $!";
	{
		local $/;
		$page_source = decode_charset( <$fh> );
	}
	close( $fh );

	#-----------------------------------------------------------------------
	#	remove crap before <!DOCTYPE
	#-----------------------------------------------------------------------

	$page_source =~ s/^.*<!DOCTYPE/<!DOCTYPE/s;

	#-----------------------------------------------------------------------
	#	remove windows style line ends
	#-----------------------------------------------------------------------

	$page_source =~ s![^\S\n]+\n!\n!smg;

	die if $page_source =~ /\r\n/;

	#-----------------------------------------------------------------------
	#	remove <!--! xxx !--> style comments
	#-----------------------------------------------------------------------

	$page_source =~
		s#^ (.*?) \s* <!--! \s+ (?: -? [^-]+ )* \s+ !--> [^\S\n]* \n #
		$1 ? "$1\n" : "" #xmge;

	die if $page_source =~ m# <!--! #x;

	#-----------------------------------------------------------------------
	#	change content type
	#-----------------------------------------------------------------------

	$page_source =~ s!
			\b content="text/html;\s*charset=iso-8859-1"
			!content="text/html;\x20charset=UTF-8"!sx;

	#-----------------------------------------------------------------------
	#	destination relative path
	#-----------------------------------------------------------------------

	$DestUri = &patch( $DestUriExpression );

	unless( $DestUri ){

		&Error( "GenPage: patch('$DestUriExpression') is empty." );
		return undef;
	}

	#-----------------------------------------------------------------------
	#	destination path (absolute or relative to current directory)
	#-----------------------------------------------------------------------

	my $dest_file = "$DestinationBase/$DestUri";

	#-----------------------------------------------------------------------
	#	create output directory
	#-----------------------------------------------------------------------

	&create_path( $dest_file );
	return if $nErrors;

	#-----------------------------------------------------------------------
	#	output to temp file
	#-----------------------------------------------------------------------

	my $temp_file = "$dest_file.$$.tmp";

	unless( open( OUT, '>:utf8', $temp_file ) ){

		&Error( "create '$temp_file': $!." );
		unlink( $temp_file );
		return;
	}

	print OUT &patch( $page_source );
	close( OUT );

	&rename_if_changed( $temp_file, $dest_file );
}

#-------------------------------------------------------------------------------
#
#	Patcher
#
#-------------------------------------------------------------------------------

sub patch{

	return (shift//'') =~
		s! [\$] ( [\w.]* ) [{] ( \w+ ) [}] !&patch2( $1,$2 )!xger;
}

sub patch2{

	my( $funcs, $var ) = @_;

	my $s = $CurrentData{ $var } // '';

	#-----------------------------------------------------------------------
	#	apply functions from right to left
	#-----------------------------------------------------------------------

	foreach my $func( reverse split( /[.]/, $funcs ) ){

		next unless $func;

		#---------------------------------------------------------------
		#
		#		predefined functions
		#
		#---------------------------------------------------------------

		#---------------------------------------------------------------
		#	recursive patch
		#---------------------------------------------------------------

		if( $func eq 'patch' ){

			$s = patch( $s );
			next;
		}

		#---------------------------------------------------------------
		#	normalise argument to lowercase [a-z_] id
		#---------------------------------------------------------------

		if( $func eq 'id' ){

			$s = lc( umlaut2ascii( $s ));
			$s =~ s!\W+!_!g;
			$s =~ s!^_+!!;
			$s =~ s!_+$!!;
			$s =~ s!_+!_!g;
			next;
		}

		#---------------------------------------------------------------
		#	remove last file extension
		#---------------------------------------------------------------

		if( $func eq 'noext' ){

			$s =~ s! [.] [^./]+ $!!x;
			next;
		}

		#---------------------------------------------------------------
		#	convert to lower case
		#---------------------------------------------------------------

		if( $func eq 'lc' ){

			$s = lc( $s );
			next;
		}

		#---------------------------------------------------------------
		#	convert to upper case
		#---------------------------------------------------------------

		if( $func eq 'uc' ){

			$s = uc( $s );
			next;
		}

		#---------------------------------------------------------------
		#	convert to title case
		#---------------------------------------------------------------

		if( $func eq 'tc' ){

			$s = uc( substr( $s, 0, 1 )) . lc( substr( $s, 1 ));
			next;
		}

		#---------------------------------------------------------------
		#	convert to integer
		#---------------------------------------------------------------

		if( $func eq 'int' ){

			$s = int( $s || 0 );
			next;
		}

		#---------------------------------------------------------------
		#	count blank separated words
		#---------------------------------------------------------------

		if( $func eq 'count' ){

			$s = scalar( grep( /\S/, split( /\s+/, $s )));
			next;
		}

		#---------------------------------------------------------------
		#	xml entity encode
		#---------------------------------------------------------------

		if( $func eq 'html' ){

			$s =~ s!&!&amp;!g;
			$s =~ s!<!&lt;!g;
			$s =~ s!>!&gt;!g;
			$s =~ s!"!&quot;!g;
			$s =~ s!'!&apos;!g;
			next;
		}

		#---------------------------------------------------------------
		#	convert space to nonbreak space
		#---------------------------------------------------------------

		if( $func eq 'nobrk' ){

			$s =~ s!\s!\xa0!g;
			next;
		}

		#---------------------------------------------------------------
		#	create sequence of paragraphs
		#---------------------------------------------------------------

		if( $func eq 'para' ){

			#-------------------------------------------------------
			#	trim and split at double newlines to paras
			#-------------------------------------------------------

			$s =~ s/^\s+//;
			$s =~ s/\s+$//;

			my @segments = split( /\s*\n\s*\n\s*/, $s );
			my @paras;

			foreach my $segment( @segments ){

				#-----------------------------------------------
				#	trim $segment
				#-----------------------------------------------

				$segment =~ s!\s+\n!\n!g;
				$segment =~ s!\s\s+!\x20!g;

				next unless $segment =~ /\S/;

				push( @paras, "<p>\n$segment\n</p>\n" );
			}

			$s = join( '', @paras );
			next;
		}

		#---------------------------------------------------------------
		#	make link relative to destination uri
		#---------------------------------------------------------------

		if( $func eq 'rlink' ){

			$s = &relative( $s, $DestUri );
			next;
		}

		#---------------------------------------------------------------
		#	make src/href links in arg relative to destination uri
		#---------------------------------------------------------------

		if( $func eq 'rlinks' || $func eq 'rcode' && $s =~ /<td>/ ){

			$s =~ s! \b (src|href) = ["'] ( [^"'<>]+ ) ["'] !
				$1 . '="' . relative( $2, $DestUri ). '"' !xeg;
			next;
		}

		#---------------------------------------------------------------
		#	make '{'uri'}' links relative to destination uri
		#---------------------------------------------------------------

		if( $func eq 'rcode' ){

			$s =~ s! [{] ( [^}]+ ) [}] !
				&relative( $1, $DestUri ) !xge;
			next;
		}

		#---------------------------------------------------------------
		#	returns german plural suffix "en" if arg != 1
		#---------------------------------------------------------------

		if( $func eq 'plural_en' ){

			$s = ( $s==1 ? '' : 'en' );
			next;
		}

		#---------------------------------------------------------------
		#	returns german plural suffix "er" if arg != 1
		#---------------------------------------------------------------

		if( $func eq 'plural_er' ){

			$s = ( $s==1 ? '' : 'er' );
			next;
		}

		#---------------------------------------------------------------
		#	convert month number to german month name
		#---------------------------------------------------------------

		if( $func eq 'monat' ){

			$s = ( '',
				'Januar', 'Februar', 'März', 'April',
				'Mai', 'Juni', 'Juli', 'August',
				'September', 'Oktober', 'November', 'Dezember'
			)[ $s ] // '';
			next;
		}

		#---------------------------------------------------------------
		#	copyright: year or firstyear..year
		#---------------------------------------------------------------

		if( $func eq 'years' ){

			my $YYYY = $Macro{ 'YYYY' };
			$s = ($s == $YYYY ? $YYYY : "$s - $YYYY");
			next;
		}

		#---------------------------------------------------------------
		#
		#		use macro definitions
		#
		#---------------------------------------------------------------

		#---------------------------------------------------------------
		#	simple translation macro
		#---------------------------------------------------------------

		if( defined $Macro{ "${func}_${s}" } ){

			$s = $Macro{ "${func}_${s}" };
			next;
		}

		#---------------------------------------------------------------
		#	simple translation macro default value
		#---------------------------------------------------------------

		if( defined $Macro{ "${func}_default" } ){

			$s = $Macro{ "${func}_default" };
			next;
		}

		#---------------------------------------------------------------
		#	plain macro expansion
		#---------------------------------------------------------------

		if( defined $Macro{ $func } ){

			local $ActiveItem  = $CurrentItem;
			local $CurrentItem = '';
			local %CurrentData = %Macro;

			my $count = 0;
			my @result;
			my @s = split( /\s+/, $s );
			my $m = $func;

			#-------------------------------------------------------
			#	iterate over group members
			#-------------------------------------------------------

			for( my $i=0; $i<@s; $i++ ){

				$CurrentItem = $s[ $i ];

				#-----------------------------------------------
				#	add item values to dataset
				#-----------------------------------------------

				foreach my $field( keys %FieldDefined ){

		$CurrentData{ $field } = $Database{ $CurrentItem, $field };
				}

				#-----------------------------------------------
				#	Element
				#-----------------------------------------------

				my $form;

				#-----------------------------------------------
				#	current = me
				#-----------------------------------------------

				$form = $Macro{ "${m}_self" }
					if $s[ $i ] eq $ActiveItem;

				#-----------------------------------------------
				#	empty if current = me and only 1 member
				#-----------------------------------------------

				$form = ''
					if @s==1 && defined $form;

				#-----------------------------------------------
				#	siblings of me
				#-----------------------------------------------

				$form = $Macro{ "${m}_next" }
					if $i>0 && ($s[$i-1]//'') eq $ActiveItem;

				$form = $Macro{ "${m}_prev" }
					if ($s[ $i+1 ] // '') eq $ActiveItem;

				#-----------------------------------------------
				#	default value
				#-----------------------------------------------

				$form = $Macro{ "${m}" } if !defined $form;

				my $res = &patch( $form );
				next if $res eq '';

				#-----------------------------------------------
				#	prefix / infix
				#-----------------------------------------------

				if( !@result ){

					$form = $Macro{ "${m}_prefix" };

				} else {
					$form = $Macro{ "${m}_infix${count }" };

					if( defined $form ){
						$count = 0;
					}else{
						$form=$Macro{ "${m}_infix" };
					}
				}

				push( @result, &patch( $form ) );

				#-----------------------------------------------
				#	item value
				#-----------------------------------------------

				push( @result, $res );
				$count++;
			}

			#-------------------------------------------------------
			#	append suffix unless no data
			#-------------------------------------------------------

			push( @result, &patch( $Macro{ "${m}_suffix" }) )
				if @result;

			$s = join( '', @result );
			next;
		}

		#---------------------------------------------------------------
		#	read long text from file
		#---------------------------------------------------------------

		if( $func =~ /^ text_ ( \w+ ) $/x ){

			&load_texts;
			$s = $LongTexts{ "$1:" . lc( $s )} // "";
			next;
		}

		#---------------------------------------------------------------
		#	default
		#---------------------------------------------------------------

		&Error( "funktion '$func' is not defined." );
	}

	return $s;
}

#-------------------------------------------------------------------------------
#
#	load named (long) texts
#
#-------------------------------------------------------------------------------

my $texts_are_loaded;

sub load_texts{

	#-----------------------------------------------------------------------
	#	read only once
	#-----------------------------------------------------------------------

	return if $texts_are_loaded;

	$texts_are_loaded = 1;

	#-----------------------------------------------------------------------
	#	iterate over text directories
	#-----------------------------------------------------------------------

	my %tag_defined;

	foreach my $dir( @TextSourceDirectories ){

		#---------------------------------------------------------------
		#	open directory
		#---------------------------------------------------------------

		unless( opendir( TXTDIR, $dir ) ){

			&Error( "opendir '$dir': $!" );
			next;
		}

		#---------------------------------------------------------------
		#	iterate over directory entries
		#---------------------------------------------------------------

		foreach my $entry( sort readdir( TXTDIR ) ){

			#-------------------------------------------------------
			#	process text files only
			#-------------------------------------------------------

			next unless $entry =~ m! [.]txt $!x;

			#-------------------------------------------------------
			#	check file access
			#-------------------------------------------------------

			my $file = "$dir/$entry";

			unless( -f $file && -r $file ){

				&Error( "$file: is not a readable file." );
				next;
			}

			#-------------------------------------------------------
			#	open file
			#-------------------------------------------------------

			unless( open( TXT, '<:raw', $file ) ){

				&Error( "load_texts: openread '$file': $!" );
				next;
			}

			warn "$file ...\n" if $Verbosity;

			#-------------------------------------------------------
			#	process file
			#-------------------------------------------------------

			my $tag;
			my @lines;

			local $File = $file;
			local $Lino = 0;

			while( <TXT> ){

				++$Lino;

				$_ = decode_charset( $_ =~ s!\s+$!!r );

				#----------------------------------------------
				#	gather lines
				#----------------------------------------------

				unless( /^@/ ){

					push( @lines, $_ ) if $tag;
					next;
				}

				#----------------------------------------------
				#	store gathered lines to LongTexts entry
				#----------------------------------------------

				$LongTexts{ $tag } = join( "\n", @lines )
								if $tag;

				#----------------------------------------------
				#	clear data
				#----------------------------------------------

				$tag   = undef;
				@lines = ();

				#----------------------------------------------
				#	next key?
				#----------------------------------------------

				unless( m! [@] ( \w+ : \w+ (?: [.-] \w+)* ) [@] !x ){

					#--------------------------------------
					#	end marker
					#--------------------------------------

					next if $_ eq '@';

					&Error( "Invalid line '$_'" );
					next;
				}

				#----------------------------------------------
				#	next key.
				#----------------------------------------------

				unless( $tag_defined{ $1 } ){

					$tag   = lc( $1 );
					@lines = ();

					$tag_defined{ $tag } = "$File:$Lino";
					next;
				}

				#----------------------------------------------
				#	multiple defined key?
				#----------------------------------------------

				&Error( "Text '$1' multiple defined:\n" .
					"\t$tag_defined{$1}\n" .
					"\t$File:$LineContent" );

				next;
			}

			warn "$file ... $Lino lines.\n" if $Verbosity;

			#-------------------------------------------------------
			#	close file
			#-------------------------------------------------------

			close( TXT );

			#-------------------------------------------------------
			#	tag still open at end of file
			#-------------------------------------------------------

			$LongTexts{ $tag } = join( "\n", @lines ) if $tag;
		}

		#---------------------------------------------------------------
		#	close directory
		#---------------------------------------------------------------

		closedir( TXTDIR );
	}
}

#-------------------------------------------------------------------------------
#
#	auxiliary filename functions
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#	extract directory name
#-------------------------------------------------------------------------------

sub dirname{

	return shift =~ s! [/\\]? [^/\\]* $!!xr;
}

#-------------------------------------------------------------------------------
#	make uri relative to a base uri
#-------------------------------------------------------------------------------

sub relative{

	my( $uri, $baseUri ) = @_;

	#-----------------------------------------------------------------------
	#	split uris to segments
	#-----------------------------------------------------------------------

	my @uri  = split( m!/+!, $uri );
	my @base = split( m!/+!, $baseUri );

	#-----------------------------------------------------------------------
	#	remove "file" from base
	#-----------------------------------------------------------------------

	pop( @base );

	#-----------------------------------------------------------------------
	#	remove common leading segments
	#-----------------------------------------------------------------------

	while( @base && @uri && ( $base[ 0 ] eq $uri[ 0 ] ) ){

		shift( @base );
		shift( @uri  );
	}

	#-----------------------------------------------------------------------
	#	use '..' to reach common root
	#-----------------------------------------------------------------------

	while( @base ){

		pop( @base );
		unshift( @uri, '..' );
	}

	return join( '/', @uri );
}

#-------------------------------------------------------------------------------
#	make file name absolute
#-------------------------------------------------------------------------------

sub absname{

	my( $name, $base ) = @_;

	#-----------------------------------------------------------------------
	#	already absolute
	#-----------------------------------------------------------------------

	return $name if $name =~ m!^ (?: [A-Z]:\\ | \\ | / ) !x;

	#-----------------------------------------------------------------------
	#	default base
	#-----------------------------------------------------------------------

	$base = $BaseDir unless $base;

	#-----------------------------------------------------------------------
	#	shortcuts
	#-----------------------------------------------------------------------

	return $base if $name eq '.';
	return $base if $name eq '';

	#-----------------------------------------------------------------------
	#	make absolute
	#-----------------------------------------------------------------------

	return "$base/$name";
}

#-------------------------------------------------------------------------------
#
#	auxiliary file functions
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#	rename file if tempfile has different content
#-------------------------------------------------------------------------------

sub rename_if_changed{

	my( $temp_file, $dest_file ) = @_;

	#-----------------------------------------------------------------------
	#	if dest file exists
	#-----------------------------------------------------------------------

	if( -f $dest_file ){

		#---------------------------------------------------------------
		#	read temp file
		#---------------------------------------------------------------

		my $temp_data;

		if( open( my $fh, '<:raw', $temp_file ) ){

			local $/;
			$temp_data = <$fh>;
			close( $fh );
		} else {

			&Fatal( "openread '$temp_file': $!" );
			die;
		}

		#---------------------------------------------------------------
		#	read current destination file content
		#---------------------------------------------------------------

		my $dest_data;

		if( open( my $fh, '<:raw', $dest_file ) ){

			local $/;
			$dest_data = <$fh>;
			close( $fh );
		} else {

			&Fatal( "openread '$dest_file': $!" );
			die;
		}

		#---------------------------------------------------------------
		#	same data: keep dest file and remove temp file
		#---------------------------------------------------------------

		if( $temp_data eq $dest_data ){

			warn "$dest_file unchanged.\n" if $Verbosity >= 1;

			unlink( $temp_file );
			return;
		}

		#---------------------------------------------------------------
		#	data not same: remove dest file
		#---------------------------------------------------------------

		warn "$dest_file changed.\n" if $Verbosity >= 1;

		unlink( $dest_file );

	} else {

		warn "$dest_file new.\n" if $Verbosity >= 1;
	}

	#-----------------------------------------------------------------------
	#	same data: keep dest file and remove temp file
	#-----------------------------------------------------------------------

	unless( rename( $temp_file, $dest_file ) ){

		&Error( "rename '$temp_file', '$dest_file': $!" );
		return;
	}
}

#-------------------------------------------------------------------------------
#
#	auxiliary directory functions
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#	rekursively create directory
#-------------------------------------------------------------------------------

sub create_path{

	my @elements = split( m! / !x, shift );

	for( my $i = 1; $i < @elements - 1; $i++ ){

		my $path = join( '/', @elements[ 0..$i ] );
		next if -d $path;

		warn "creating '$path'...\n" if $Verbosity >= 2;

		unless( mkdir( $path, 0755 ) ){

			&Error( "mkdir '$path': $!" );
			return;
		}
	}
}

#-------------------------------------------------------------------------------
#
#	auxiliary string functions
#
#-------------------------------------------------------------------------------

sub decode_charset{

	my( $s ) = @_;

	#-----------------------------------------------------------------------
	#	try utf-8
	#-----------------------------------------------------------------------

	my $decoded = Encode::decode( 'utf-8', $s );

	#-----------------------------------------------------------------------
	#	if failed, try cp1252 aka Microsoft ISO
	#-----------------------------------------------------------------------

	if( index( $decoded, chr(0xfffd) ) >= 0 ){

		$decoded = Encode::decode( 'cp1252', $s );
	}

	return $decoded unless index( $decoded, chr(0xfffd) ) >= 0;

	&Fatal( "Unknown character encoding." );
}

#-------------------------------------------------------------------------------
#	create id consisting of ascii chars and digits and underscore only
#-------------------------------------------------------------------------------

sub genid {
	my $s = lc( umlaut2ascii( shift ));
	$s =~ s!\W+!_!g;
	$s =~ s!^_+!!;
	$s =~ s!_+$!!;
	$s =~ s!_+!_!g;
	return $s;
}

#-------------------------------------------------------------------------------
#	remove common (german) umlauts
#-------------------------------------------------------------------------------

sub umlaut2ascii{

	my $s = shift;

	$s =~ s/Ä/Ae/g;
	$s =~ s/Ö/Oe/g;
	$s =~ s/Ü/Ue/g;
	$s =~ s/ä/ae/g;
	$s =~ s/ö/oe/g;
	$s =~ s/ü/ue/g;
	$s =~ s/ß/ss/g;

	return $s;
}

#-------------------------------------------------------------------------------
#       $Id: bilderwichtel.pl,v 1.22 2023/02/14 21:26:35 wolf Exp $
#-------------------------------------------------------------------------------
